// Copyright 2018 The Oppia Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/**
 * @fileoverview Service to manage what audio translations are currently
 * being played or paused.
 */

import { Injectable } from '@angular/core';
import { downgradeInjectable } from '@angular/upgrade/static';

import { Voiceover } from
  'domain/exploration/voiceover.model';

export interface AudioTranslations {
  [languageCode: string]: Voiceover;
}

@Injectable({
  providedIn: 'root'
})
export class AudioTranslationManagerService {
  // Audio translations for the main content of a card at the top.
  _contentAudioTranslations: AudioTranslations = {};
  _contentHtmlForAutogeneratedAudio: string = '';
  _currentPrimaryComponentName: string = '';

  // Audio translations for everything other than the main content, such as
  // those for feedback, hints, and solutions.
  _secondaryAudioTranslations: AudioTranslations = {};
  _secondaryHtmlForAutogeneratedAudio: string = '';
  _currentSecondaryComponentName: string = '';

  // Audio translations that will play back after the main content audio has
  // played. Useful for reading out loud multiple choices.
  _sequentialAudioTranslations: AudioTranslations = {};
  _sequentialHtmlForAutogeneratedAudio: string = '';
  _currentSequentialComponentName: string = '';

  setContentAudioTranslations(
      audioTranslations: AudioTranslations, html: string,
      componentName: string): void {
    this._contentAudioTranslations = audioTranslations;
    this._contentHtmlForAutogeneratedAudio = html;
    this._currentPrimaryComponentName = componentName;

    // Clear previously set sequential audio translations for those lessons
    // that do not have sequential audio, as we don't want the sequential
    // audio of other lessons to leak.
    this.clearSequentialAudioTranslations();
  }

  setSecondaryAudioTranslations(
      audioTranslations: AudioTranslations, html: string,
      componentName: string): void {
    this._secondaryAudioTranslations = audioTranslations;
    this._secondaryHtmlForAutogeneratedAudio = html;
    this._currentSecondaryComponentName = componentName;

    // Clear previously set sequential audio translations as we don't want
    // to play them after the secondary audio translations.
    this.clearSequentialAudioTranslations();
  }

  // This needs to be called *after* setContentAudioTranslations; e.g. inside
  // ngOnInit() of the component.
  setSequentialAudioTranslations(
      audioTranslations: AudioTranslations,
      html: string,
      componentName: string
  ): void {
    this._sequentialAudioTranslations = audioTranslations;
    this._sequentialHtmlForAutogeneratedAudio = html;
    this._currentSequentialComponentName = componentName;
  }

  clearSecondaryAudioTranslations(): void {
    this._secondaryAudioTranslations = {};
    this._secondaryHtmlForAutogeneratedAudio = '';
    this._currentSecondaryComponentName = '';
  }

  clearSequentialAudioTranslations(): void {
    this._sequentialAudioTranslations = {};
    this._sequentialHtmlForAutogeneratedAudio = '';
    this._currentSequentialComponentName = '';
  }

  // To make sure that audio does not repeat when there
  // is no audio for the seconday text, we return
  // an empty object if there is secondary HTML.
  getCurrentAudioTranslations(): AudioTranslations {
    if (Object.keys(this._secondaryAudioTranslations).length !== 0 ||
      this._secondaryHtmlForAutogeneratedAudio !== '') {
      return this._secondaryAudioTranslations;
    }
    return this._contentAudioTranslations;
  }

  getCurrentHtmlForAutogeneratedAudio(): string {
    if (this._secondaryHtmlForAutogeneratedAudio !== '') {
      return this._secondaryHtmlForAutogeneratedAudio;
    }
    return this._contentHtmlForAutogeneratedAudio;
  }

  getCurrentHtmlForAutogeneratedSequentialAudio(): string {
    return this._sequentialHtmlForAutogeneratedAudio;
  }

  getCurrentComponentName(): string {
    if (this._currentSecondaryComponentName !== '') {
      return this._currentSecondaryComponentName;
    }
    return this._currentPrimaryComponentName;
  }

  cleanUpHTMLforVoiceover(html: string): string {
    // If the labels are in html format, remove the tags and leave the
    // content only.
    const cleanChoiceLabel = html.replace(/<[^>]+>/g, '');

    // Add a stop for the voiceover with a dot. Useful for multiple choices.
    return cleanChoiceLabel + '. ';
  }
}

angular.module('oppia').factory(
  'AudioTranslationManagerService',
  downgradeInjectable(AudioTranslationManagerService));
